/*
 *  Gmail2Pager
 *
 *  Copyright (C) 2011-2014 Clive Cooper.
 *
 *  Gmail2Pager is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  Gmail2Pager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Clive Cooper,
 *
 */

#include <QtGui>
#include <QApplication>
#include <QSslSocket>
//#include <QDebug>
#include "mainwin.h"


const QString VERSION = "2013.11.08-1";
#define ENDQTY 13
const QString SUBEND[ENDQTY] = {"From:", "To:", "References:", "Date:", "MIME-Version:", "X-MaxCode-Template:", "Content-Type:",
			    "X-PHP-Originating-Script:", "Reply-To:", "Supposedly-From:", "Message-Id:", "X-AntiAbuse:", "Bounces-to:"};
#define QSn QString::number


MainWin::MainWin( QWidget * parent, Qt::WFlags f) 
	: QMainWindow(parent, f)
{
  setupUi(this);
  running = false;
  userPath = QDir::homePath();
  userDir = QDir::homePath() + "/.gascop";
  pagerIDdb = new QStringList();
  gascopdb = new QStringList();
  check = new QTimer(this);
  connect(check , SIGNAL(timeout()), this, SLOT(checkTimeout()));
  socket = new QSslSocket(this);
  connect(socket, SIGNAL(readyRead()), this, SLOT(readyReadSocket()));
  connect(pbStart , SIGNAL(clicked()), this, SLOT(startClicked()));
  connect(pbStop , SIGNAL(clicked()), this, SLOT(stopClicked()));
  connect(socket , SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
  connect(sbcheckPeriod , SIGNAL(valueChanged(int)), this, SLOT(checkPeriodChanged(int)));
}

// This allows me to do seconds, milliseconds and microseconds sleeping : gotsleep::msleep(100); for 100 milliseconds
class gotoSleep : public QThread
{
  public:
    static void sleep(unsigned long secs)
    {
      QThread::sleep(secs);
    }
    static void msleep(unsigned long msecs)
    {
      QThread::msleep(msecs);
    }
    static void usleep(unsigned long usecs)
    {
      QThread::usleep(usecs);
    }
};

void MainWin::showEvent(QShowEvent *)
{
  if(running == false)
  {
    running = true;
    lVersion->setText("Version: " + VERSION);
    readSettings();
    stopChecking = false;
    isChecking = false;
    pbStop->setEnabled(false);
    totalPages = 0;
    emailCount = 0;
    emailsOnSvr = -1;
    dataIn = "";
    countDown = sbcheckPeriod->value() * 60;
    check->setInterval(1000);
  }
}

void MainWin::moveEvent(QMoveEvent *)
{
  formleft = x();
  formtop = y();
}

void MainWin::closeEvent(QCloseEvent *)
{
  stopClicked();
  writeSettings();
}

void MainWin::startClicked()
{
  pbStop->setEnabled(true);
  pbStart->setEnabled(false);
  countDown = 1;
  checkTimeout();
}

void MainWin::stopClicked()
{
  pbStop->setEnabled(false);
  if(isChecking)
    stopChecking = true;
  else
  {
    check->stop();
    pbStart->setEnabled(true);
    status->setText("Stopped");
  }
}

void MainWin::checkPeriodChanged(int val)
{
  countDown = val * 60;
}

void MainWin::checkTimeout()
{
  status->setText("Checking in " + QSn(countDown--) + " seconds");
  if(countDown <= 0)
  {
    countDown = sbcheckPeriod->value() * 60;
    check->stop();
    isChecking = true;
    subjects.clear();
    dataIn = "";
    socket->connectToHostEncrypted(leServer->text(), 995);
  }
}

void MainWin::readyReadSocket()
{
  dataIn += socket->readAll();
  if(dataIn.left(32) == "+OK Gpop ready for requests from")   // Got a connection
  {
    dataIn = "";
    status->setText("Logging in.");
    socket->write("USER recent:" + leUsername->text().toUtf8() + "\r\n"); // send the username
  }
  else if(dataIn == "+OK send PASS\r\n")                      // Send the password
  {
    dataIn = "";
    socket->write("PASS " + lePassword->text().toUtf8() + "\r\n");
  }
  else if(dataIn == "+OK Welcome.\r\n")                       // Auth OK so do a STAT and get how many (recent) emails available
  {
    dataIn = "";
    status->setText("Checking for emails");
    socket->write("STAT\r\n");
  }
  else if(dataIn == "+OK marked for deletion\r\n")            // Accepting deletion
  {
    dataIn = "";
    emailCount--;
    if(emailCount > 0)
      socket->write("RETR " + QSn(emailCount).toUtf8() + "\r\n");
    else
      socket->write("QUIT\r\n");
  }
  else if(dataIn.right(5) == "\r\n.\r\n")                     // This is what is at the end of emails
  {
    int start = dataIn.indexOf("\r\nSubject:"); // GASCOP:BT EasyReach 77C:Clive:This is a longer message than usual because it is a test.
    int end = getSubjectEnd(start);
    QString sub = dataIn.mid(start + 11, end - (start + 11)).trimmed();
    sub = sub.remove("\r\n");
    if((sub.left(7) == "GASCOP:") && (sub.count(':') > 2))  // To Qualify it must start with GASCOP: and have 3 or more colons
    {
      subjects.append(sub);
      status->setText("Found " + QSn(subjects.count()) + " pager emails.");
      totalPages++;
      total->setText("Total Pages Found=" + QSn(totalPages));
      socket->write("DELE " + QSn(emailCount).toUtf8() + "\r\n"); // delete this email now we have it
      emailsOnSvr--;
      dataIn = "";
      return;
    }
    dataIn = "";
    emailCount--;
    if(emailCount > 0)
      socket->write("RETR " + QSn(emailCount).toUtf8() + "\r\n"); // Ask for the next email
    else
      socket->write("QUIT\r\n");
  }
  else if(dataIn == "+OK Farewell.\r\n")                    // disconnected
  {
    dataIn = "";
    status->setText("Connection Closed.");
  }
  else if(dataIn.left(21) == "+OK message follows\r\n")
    return;
  else if((dataIn.left(4) == "+OK ") && (dataIn.mid(4,1) >= "0") && (dataIn.mid(4,1) <= "9")) // "+OK 15 505865\r\n"
  {
    QStringList lst = dataIn.split(" ");
    dataIn = "";
    emailCount = lst.at(1).toInt();
    if((emailCount != emailsOnSvr) && (emailCount > 0))
      socket->write("RETR " + lst.at(1).toUtf8() + "\r\n");
    else
      socket->write("QUIT\r\n");
    if(emailsOnSvr < 0)
      emailsOnSvr = emailCount;
  }
}

// Searches for the end of the subject line looking for all the kinds of fields that may follow a Subject: field
int MainWin::getSubjectEnd(int from)
{
  int end = 1000000;
  for(int x = 0; x < ENDQTY; x++)
  {
    int y = dataIn.indexOf(SUBEND[x], (from + 11), Qt::CaseInsensitive);
    if((y < end) && (y > 0))
      end = y;
  }
  return(end);
}

void MainWin::socketDisconnected()
{
  if(stopChecking == false)
    check->start();
  else
  {
    stopChecking = false;
    status->setText("Stopped");
    pbStart->setEnabled(true);
  }
  isChecking = false;
  if(subjects.count())
    storePages();
}

// get the pager details from the pager ID database and build a pager message entry for the pages database
void MainWin::storePages()
{
// A pager message consists of    GASCOP:PAGER_ID:FROM:MESSAGE
  readTextFile(userDir + "/pagerid.db", pagerIDdb);
  getPagerIDElements();
  //subjects << "GASCOP:BT EasyReach 77C:Joe Crow:Another test message";   // DEBUG ONLY
  gascopdb->clear();
  for(int x = 0; x < subjects.count(); x++)   // Iterate through all email/pages and find what pager the message is for and add it to the msgs database
  {
    QStringList pm = subjects.at(x).split(":", QString::KeepEmptyParts);
    if(pm.count() < 3)                              // Not enough elements so throw it out
      continue;
    while(pm.count() > 4)                           // This happens if colons are used in the message and this preserves them
    {
      pm.replace(3, pm.at(3) + ":" + pm.at(4));
      pm.removeAt(4);
    }
    int index = -1;
    for(int y = 0; y < dbPagerIDs.count(); y++)
    {
      if(pm.at(1) == dbPagerIDs.at(y).name)
      {
        index = y;
        break;
      }
    }
    if(index >= 0)
    {
      QString ric = QSn(dbPagerIDs.at(index).ric);
      while(ric.length() < 7)
        ric = "0" + ric;
      QString baud = QSn(dbPagerIDs.at(index).baud);
      while(baud.length() < 4)
        baud = "0" + baud;
      QString dbEntry;
      if((pm.at(3) == "") || (dbPagerIDs.at(index).numeric == 1))
      {
        dbEntry = "0" + QSn(timeToInt()) + "000000000" + QSn(dbPagerIDs.at(index).tx) + baud + QSn(dbPagerIDs.at(index).numeric) +
                        QSn(dbPagerIDs.at(index).funcBits) + ric + pm.at(3);
      }
      else
        dbEntry = "0" + QSn(timeToInt()) + "000000000" + QSn(dbPagerIDs.at(index).tx) + baud + QSn(dbPagerIDs.at(index).numeric) +
                        QSn(dbPagerIDs.at(index).funcBits) + ric + pm.at(3) + " -From:" + pm.at(2);
      gascopdb->append(dbEntry);
    }
  }
  if(gascopdb->count() > 0)
  {
    setFileLock();
    appendTextFile(userDir + "/gascop.db", gascopdb);
    QFile::remove(userDir + "/gascop.db.lck");
  }
}

// Supplied with just the filename (like... gascop.db) this will create a lock file, if there is a lock file it waits 3 seconds and deletes it
void MainWin::setFileLock()
{
  uint t = timeToInt() + 2;                                 // Wait for up to 2 seconds for the lock to expire
  while(QFile::exists(userDir + "/gascop.db.lck"))   // Loop while we wait
  {
    gotoSleep::msleep(200);                                 // This while loop can use a lot of cpu so we sleep for a good part of the time
    QCoreApplication::processEvents(QEventLoop::AllEvents); // Keep the GUI responsive
    if(timeToInt() >= t)                                    // Must be a stale lock to get into here: the 2 seconds is up
    {
      QFile::remove(userDir + "/gascop.db.lck");     // Delete the lock
      break;                                                // Break out of the loop
    }
  }
  QFile::copy(":/files/gascop.db.lck", userDir + "/gascop.db.lck"); // write a new lock file (Remember it expires in 2 seconds!)
}

// Returns the elements of the entries in the pager ID database
void MainWin::getPagerIDElements()
{
  dbPagerIDs.clear();
  int index = 0;
  while(index < pagerIDdb->count())
  {
    dbPagerID.ric = pagerIDdb->at(index).left(7).toInt();
    dbPagerID.funcBits = pagerIDdb->at(index).mid(7, 1).toInt();
    dbPagerID.numeric = pagerIDdb->at(index).mid(8, 1).toInt();
    dbPagerID.baud = pagerIDdb->at(index).mid(9, 4).toInt();
    dbPagerID.tx = pagerIDdb->at(index).mid(13, 1).toInt();
    dbPagerID.fx = pagerIDdb->at(index).mid(14, 9);
    dbPagerID.msgLen = pagerIDdb->at(index).mid(23, 3).toInt();
    dbPagerID.name = pagerIDdb->at(index++).mid(26);
    dbPagerIDs.append(dbPagerID);
  }
}

// Get UNIX time as a unsigned long integer (should work on all platforms)
uint MainWin::timeToInt()
{
  return(QDateTime::currentDateTime().toTime_t());
}

// This Encrypts and Decrypts a string
QString MainWin::dencrypt(QString dataIn, QString key)
{
  int x, y, t;
  char a, b;
  QByteArray u, v, m;
  u = dataIn.toUtf8();
  v = key.toUtf8();
  t = v.size() - 1;
  y = 0;
  for(x = 0; x < dataIn.size(); x++)
  {
    a = (u[x] & 0x0f) ^ (v[y] & 0x0f);
    b = (u[x] & 0xf0) + a;
    u[x] = b;
    y++;
    if (y > t)
      y = 0;
  }
  return QString(u);
}

// Reads a text file line by line into a string list supplied by the caller
bool MainWin::readTextFile(QString fn, QStringList* sl)
{
  bool res;
  res = false;
  sl->clear();
  QFile file(fn);
  if(file.exists())
  {
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
      return(res);
    QTextStream lineIn(&file);
    while (!lineIn.atEnd())
      {
        QString line = lineIn.readLine();
        sl->append(line);
      }
    file.close();
    res = true;
  }
  return(res);
}

// Writes the contents of a string list to a file adding a \n at the end of each line
bool MainWin::writeTextFile(QString fn, QStringList* sl)
{
  bool res;
  res = false;
  QFile file(fn);
  if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
    return(res);
  QTextStream out(&file);
  for(int x = 0; x < sl->count(); x++)
  {
    QString remindLine = sl->at(x) + "\n";
    out << remindLine;
  }
  file.close();
  res = true;
  return(res);
}

// Append the contents of a string list to a file adding a \n at the end of each line
bool MainWin::appendTextFile(QString fn, QStringList* sl)
{
  bool res;
  res = false;
  if(sl->count() > 0)
  {
    QFile file(fn);
    if(!file.open(QIODevice::Append | QIODevice::Text))
      return(res);
    QTextStream out(&file);
    for(int x = 0; x < sl->count(); x++)
    {
      QString remindLine = sl->at(x) + "\n";
      out << remindLine;
    }
    file.close();
    res = true;
  }
  return(res);
}

void MainWin::writeSettings()
{
  QSettings settings(QSettings::IniFormat, QSettings::UserScope,"CLPC-Software", "gmail2page");
  settings.setValue("left", formleft);
  settings.setValue("top", formtop);
  settings.setValue("height", height());
  settings.setValue("width", width());
  settings.setValue("server", leServer->text());
  settings.setValue("username", leUsername->text());
  settings.setValue("password", dencrypt(lePassword->text(), "a50Bah6bsK2huaqWi7Gf"));
  settings.setValue("checkperiod", sbcheckPeriod->value());
}

void MainWin::readSettings()
{
  QSettings settings(QSettings::IniFormat, QSettings::UserScope,"CLPC-Software", "gmail2page");
  move(settings.value("left", 100).toInt(), settings.value("top", 100).toInt());
  resize(settings.value("width", 550).toInt(), settings.value("height", 400).toInt());
  leServer->setText(settings.value("server", "pop.gmail.com").toString());
  leUsername->setText(settings.value("username", "your_gmail_account@gmail.com").toString());
  lePassword->setText(dencrypt(settings.value("password", "1234").toString(), "a50Bah6bsK2huaqWi7Gf"));
  sbcheckPeriod->setValue(settings.value("checkperiod", 5).toInt());
}
