/*
 *  Gmail2Pager
 *
 *  Copyright (C) 2011-2014 Clive Cooper.
 *
 *  Gmail2Pager is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  Gmail2Pager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Clive Cooper,
 *
 */

#ifndef MAINWIN_H
#define MAINWIN_H

#include "ui_mainwin.h"

class QSslSocket;
class QTimer;

class MainWin : public QMainWindow, public Ui::MainWin
{
Q_OBJECT
  public:
    MainWin( QWidget * parent = 0, Qt::WFlags f = 0 );

  protected:
    void showEvent(QShowEvent *);
    void closeEvent(QCloseEvent *);
    void moveEvent(QMoveEvent *);

  private slots:
    void readyReadSocket();
    void startClicked();
    void stopClicked();
    void socketDisconnected();
    void checkTimeout();
    void checkPeriodChanged(int val);

  private:
    bool running;
    QString userPath;
    int formleft, formtop;
    QSslSocket * socket;
    QString dataIn;
    int emailCount;
    int emailsOnSvr;
    QStringList subjects;
    QTimer * check;
    bool stopChecking;
    bool isChecking;
    int totalPages;
    QString userDir;
    QStringList * pagerIDdb;
    QStringList * gascopdb;
    int countDown;

    struct pid
    {
      QString name;
      int ric;
      int funcBits;
      int numeric;
      int baud;
      int tx;
      QString fx;
      int msgLen;
    };
    QVector<pid> dbPagerIDs;
    pid dbPagerID;

    int getSubjectEnd(int from);
    void setFileLock();
    uint timeToInt();
    bool readTextFile(QString fn, QStringList* sl);
    bool writeTextFile(QString fn, QStringList* sl);
    bool appendTextFile(QString fn, QStringList* sl);
    void getPagerIDElements();
    void storePages();
    QString dencrypt(QString dataIn, QString key);
    void writeSettings();
    void readSettings();

};
#endif

